//  Copyright (C) 2001-2003, Evolution Robotics, Inc.
//  Any reproduction is strictly prohibited without the explicit
//  written permission of Evolution Robotics, Inc.  All rights reserved.


/**
 * @file    bump_sensor.cpp
 *
 * Test utility for cameras
 */

//#include <cstdio>
#include <signal.h>
#include <evolution/Resource.hpp>

// Globals used by signal handler
#define SENSOR_NAME         "Camera"
#define SENSOR_INTERFACE    "Evolution.ICamera"

// Convenient typedefs for Evolution types
typedef Evolution::Result              Result;
typedef Evolution::ResourceManager     ResourceManager;
typedef Evolution::ResourceConfigParser ResourceConfigParser;
typedef Evolution::IResourceContainer  IResourceContainer;
typedef Evolution::IResource           IResource;
typedef Evolution::ICamera             ICamera;
typedef Evolution::DeviceBusConfig     DeviceBusConfig;
typedef Evolution::DeviceList          DeviceList;
typedef Evolution::String              String;
typedef Evolution::Image               Image;

typedef ICamera Sensor;
typedef std::vector<Sensor*> SensorList;
typedef std::vector<const char*> StringList;

SensorList* sensors = NULL;
StringList* names = NULL;
IResourceContainer* resource_container = NULL;
ResourceManager* g_manager = NULL;
std::auto_ptr<ResourceManager> g_manager_cleanup;

void shutdown_interrupt(int signum)
{
    static bool shutting_down = false;
    signal(signum, shutdown_interrupt);
    if (signum == SIGINT)
    {
        if (shutting_down)
        {
            return;
        }
        shutting_down = true;
    }
}

void exit_handler (void)
{
    /**if (g_manager && (g_manager->is_active ()))
    {
        g_manager->deactivate ();
    }**/
}

Sensor* get_sensor(const char* sensor_id, bool print_error)
{
    // Get sensor interface
    Sensor* sensor = NULL;
    if ( resource_container->obtain_interface (0, sensor_id, 
          SENSOR_INTERFACE, (void**) &sensor) != Evolution::RESULT_SUCCESS )
    {
        if (print_error)
            std::cout << "ERROR: Resource manager failed to obtain " SENSOR_NAME 
                      << " '" << sensor_id << "' (maybe it's not setup as a " 
                      << SENSOR_NAME "?)\n";
        return NULL;
    }

    if (sensor == NULL && print_error)
        std::cout << "ERROR: " SENSOR_NAME " '" << sensor_id << "' not found.\n";

    return sensor;
}

bool get_sensor_list(SensorList* sensors, StringList* names)
{
    std::cout << "Accessing all " SENSOR_NAME "s in system.\n";

    const ResourceConfigParser* r_config = NULL;
    if (resource_container->get_configuration(0, &r_config) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Could not load Resource Configuration parser.\n";
        return false;
    }

    const ResourceConfigParser::BusList& bus_list = r_config->get_bus_list(); 
    for (ResourceConfigParser::BusList::const_iterator bus_iter = bus_list.begin();
         bus_iter != bus_list.end(); ++bus_iter)
    {
        //DeviceBusConfig* device_list = bus_iter.get_data();
        const DeviceBusConfig* device_bus = (*bus_iter);//.devices.begin(); 
        
        //device_list iterator.
        for (DeviceList::const_iterator device_iter = device_bus->devices.begin();
             device_iter != device_bus->devices.end(); ++device_iter) 
        {
            // Get the sensor's name
            String sensor_id;
            if ((*device_iter)->get_id(&sensor_id) != Evolution::RESULT_SUCCESS)
                continue;

            // Get the sensor interface
            Sensor* sensor = get_sensor(sensor_id.c_str(), false);
            if (sensor == NULL)
                continue;

            // Its valid-- release the interface for now
            resource_container->release_interface (0, sensor);  

            // Track sensor in list
            const char *name = sensor_id.c_str();
            char *copy = new char[strlen(name)+1];
            strcpy(copy, name);
            sensors->push_back(sensor);
            names->push_back(copy);
            std::cout << "    " << sensor_id.c_str() << "\n";
        }
    }

    std::cout << "\n";

    return true;
}

void test_sensors(SensorList* sensors, StringList* names,
                  int frames, double quality)
{
    std::cout << "Reading " SENSOR_NAME "s " << frames << " times.\n";

    for (int frame = 1; frame <= frames; frame++)
    {
        std::cout << "Frame " << frame << ": ";

        for (unsigned int i = 0; i < sensors->size() && names->size(); i++)
        {
            std::cout << "  " << (*names)[i] << ": ";

            const Image* image;
            if ((*sensors)[i]->get_image(0, &image) != Evolution::RESULT_SUCCESS)
            {
                std::cout << "Could not get image it took\n";
                continue;
            }

            char filename[80];
            std::sprintf(filename, "%s_%.4i.jpg", (*names)[i], frame);
            if (image->write_file(filename, quality) != Evolution::RESULT_SUCCESS) {
                std::cout << "Failed to write image\n";
                continue;
            } 

            std::cout << "Wrote image\n";
        }

        if (frame < frames) {
            std::cout << "\n";
            Evolution::Platform::millisecond_sleep(500);
        }
    }
}

int main (int argc, char** argv)
{
    // Signal Handling
    signal(SIGINT, shutdown_interrupt);
#if !defined(EVOLUTION_PLATFORM_WIN32)
    signal(SIGCHLD, shutdown_interrupt);
    signal(SIGPIPE, shutdown_interrupt);
#endif

    // Exit handling
    std::atexit (exit_handler);

    // Set up logging.
    ERSP_LOG_SET_ROOT_PRIORITY(LOG_INFO);

    //set up containers
    Result result = Evolution::RESULT_SUCCESS;
    ResourceManager* manager;
    manager = new ResourceManager (NULL, result);  // create it then globalize it...
    g_manager = manager;
    g_manager_cleanup.reset (g_manager);
    if (result != Evolution::RESULT_SUCCESS || manager == null)
    {
        std::cout << "ERROR: Failed to load resource manager.\n";
        return 1;
    }

    if (manager->get_resource_container(0, &resource_container) != Evolution::RESULT_SUCCESS )
    {
        std::cout << "ERROR: Failed to get resource container from resource manager.\n";
        return 1;
    }

    if (resource_container == NULL)
    {
        std::cout << "ERROR: Got NULL resource container from resource manager.\n";
        return 1;
    }

    Evolution::Platform::millisecond_sleep(2000);

    // Create list of sensors to test
    SensorList sensor_list;
    StringList name_list;
    sensors = &sensor_list; // Globalized for shutdown_interrupt()
    names = &name_list;

    double quality = .90;
    int frames = 5;
    if (argc <= 1)
    {
        std::cout << "Usage: [camera name]... [options].  Use for " SENSOR_NAME "s only.\n"
                  << "Options:\n"
                  << "  --quality <num>    jpeg quality\n"
                  << "  --frames  <num>    number of frames to output\n" 
                  << "\nUsing all cameras found in system at " << quality
                  << "quality for " << frames << " frames\n";

//FIXME: We want to do this even when they specity frames and quality
//but no camera names...  How do we merge that together?
        // Use all sensors in system
        if (!get_sensor_list(&sensor_list, &name_list))
            return (1);
    }

    else
    {
        // Use requested sensors
        for (int i = 1; i < argc; i++)
        {
            if (!strcmp(argv[i], "--frames"))
            {
                // Make sure the array contains another entry
                if (++i < argc)
                    frames = atoi(argv[i]);
                continue;
            }

            if (!strcmp(argv[i], "--quality"))
            {
                // Make sure the array contains another entry
                if (++i < argc)
                    quality = atoi(argv[i]);
                continue;
            }

            Sensor* sensor = get_sensor(argv[i], true);
            if (sensor) {
                const char *name = argv[i];
                char *copy = new char[strlen(name)+1];
                strcpy(copy, name);
                name_list.push_back(copy);
                sensor_list.push_back(sensor);
            }
        }
    }

    if (!sensors->size()) {
        std::cout << "ERROR: No valid " SENSOR_NAME "s found.\n";
        return (1);
    }

    test_sensors(&sensor_list, &name_list, frames, quality);

    std::cout << "Deactivating resources\n";

    // Release allocated interfaces
    if (resource_container && sensors) {
        for (unsigned int i = 0; i < sensors->size(); i++)
            resource_container->release_interface (0, (*sensors)[i]);
        sensors = NULL;
    }

    // Delete string arrays
    if (names) {
        for (unsigned int i = 0; i < names->size(); i++)
            delete [] (*names)[i];
    }
    return (0);
}
