//  Copyright (C) 2001-2003, Evolution Robotics, Inc.
//  Any reproduction is strictly prohibited without the explicit
//  written permission of Evolution Robotics, Inc.  All rights reserved.


/**
 * @file    evo-driver/tests/diff_drive.hpp
 * @author  Copyright (C) 2002, Evolution Robotics, Inc.
 *
 * Unit tests of Diff2Drive. Currently these tests are quite dependent
 * on ResourceManager, because the drive system is an aggregate
 * resource that must search for its component resources. But to avoid
 * this dependency, a large subset of ResourceManager's functionality
 * would have to be duplicated in some test IResourceContainer. It's
 * not clear that the effort is at all worthwhile.
 */

// includes for evolution motors/IRs
#include <cmath>
#include <signal.h>
#include <evolution/Resource.hpp>
#include <evolution/Util.hpp>

// includes for server
#define WIN32_LEAN_AND_MEAN
//#include <winsock2.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// globals for server
#define DEFAULT_PORT 5001
//#define DEFAULT_PROTO SOCK_STREAM // TCP
#define DEFAULT_PROTO SOCK_DGRAM //UDP

// global declarations for IR
#define SENSOR_NAME         "Range sensor"
#define SENSOR_INTERFACE    "Evolution.IRangeSensor"

using namespace std;

// Convenient typedefs for Evolution types
typedef Evolution::Result              Result;
typedef Evolution::ResourceManager     ResourceManager;
typedef Evolution::ResourceCallbackImpl ResourceCallbackImpl;
typedef Evolution::ResourceConfigParser ResourceConfigParser;
typedef Evolution::IResourceContainer  IResourceContainer;
typedef Evolution::IResource           IResource;
typedef Evolution::IDriveSystem        IDriveSystem;
typedef Evolution::CallbackId          CallbackId;
typedef Evolution::AcquirableLock      AcquirableLock;
typedef Evolution::IConditionVariable  IConditionVariable;
typedef Evolution::DeviceBusConfig     DeviceBusConfig;
typedef Evolution::DeviceList          DeviceList;
typedef Evolution::String              String;
typedef Evolution::IRangeSensor        IRangeSensor;
typedef Evolution::Timestamp           Timestamp;

// global declarations for IRs
typedef IRangeSensor Sensor;
typedef vector<Sensor*> SensorList;
typedef vector<const char*> StringList;

SensorList* sensors = NULL;
StringList* names = NULL;

// global declaratins for motors
IDriveSystem* drive = NULL;
IResourceContainer* resource_container = NULL;
ResourceManager* g_manager = NULL;
auto_ptr<ResourceManager> g_manager_cleanup;

// global declarations for server
int socket_type = DEFAULT_PROTO;
SOCKET listen_socket;
struct sockaddr_in from;
static bool shutting_down = false;

void shutdown_interrupt(int signum)
{
    signal(signum, shutdown_interrupt);
    if (signum == SIGINT)
        shutting_down = true;
}

void exit_handler (void)
{
    if (shutting_down)
    {
        /**if (g_manager && (g_manager->is_active ()))
        {
            g_manager->deactivate ();
        }**/
    }
}

void CheckResult(bool test)
{
    std::cout << (test ? "passed" : "failed") << "\n";
}

void Usage(char *progname) {
    fprintf(stderr,"Usage\n%s -p [protocol] -e [endpoint] -i [interface_server]\n",
        progname);
    fprintf(stderr,"Where:\n\tprotocol is one of TCP or UDP\n");
    fprintf(stderr,"\tendpoint is the port to listen on\n");
    fprintf(stderr,"\tinterface is the ipaddr (in dotted decimal notation)");
    fprintf(stderr," to bind to\n");
    fprintf(stderr,"Defaults are TCP,5001 and INADDR_ANY\n");
    WSACleanup();
    exit(1);
}

int sensor_init(int argc, char** argv) {
	 // Create list of sensors to test
    SensorList sensor_list;
    StringList name_list;
    sensors = &sensor_list; // Globalized for shutdown_interrupt()
    names = &name_list;

    if (argc <= 1) {
        // Use all sensors in system
        if (!get_sensor_list(&sensor_list, &name_list))
            return (1);
    }
    else {
        // Use requested sensors
        for (int i = 1; i < argc; i++) {
            Sensor* sensor = get_sensor(argv[i], true);
            if (sensor) {
                const char *name = argv[i];
                char *copy = new char[strlen(name)+1];
                strcpy(copy, name);
                name_list.push_back(copy);
                sensor_list.push_back(sensor);
            }
        }
    }

    if (!sensors->size()) {
        std::cout << "ERROR: No valid sensors found.\n";
        return (1);
    }
    else
        test_sensors(&sensor_list, &name_list, 10);
}

/* Initialize Evolution motors */
int motor_init() {
	
	// Get the interface pointer.
    std::cout << "Obtaining the drive system interface\n";
    char* interface_name = "drive";
    if (resource_container->obtain_interface (0, interface_name,
                                     IDriveSystem::INTERFACE_ID,
                                     (void**) &drive) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Resource manager failed to obtain Drived System"
                  << " '" << interface_name << "' (maybe it's not setup as a " 
                  << "Drive system?)\n";
        return (1);
    }

    if (drive == NULL) {
        std::cout << "ERROR: Drive system '" << interface_name << "' not found.\n";
        return (1);
    }

	return 0;
}

int evolution_init(int argc, char** argv, Result result, ResourceManager* manager) {


    // Exit handling
    atexit (exit_handler);

    ResourceCallbackImpl callback;

    g_manager = manager;
    g_manager_cleanup.reset (g_manager);
    if (result != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Failed to load resource manager.\n";
        return 1;
    }

    if (manager->get_resource_container (0, &resource_container) != Evolution::RESULT_SUCCESS)
    {
        std::cout << "ERROR: Failed to get resource container from resource manager.\n";
        return 1;
    }

    if (resource_container == NULL)
    {
        std::cout << "ERROR: Got NULL resource container from resource manager.\n";
        return 1;
    }

    Evolution::Platform::millisecond_sleep(2000);

    motor_init()
	sensor_init(argc, argv)
	
}



int server_init(int argc, char** argv, unsigned short port) {

	int i, retval;
	char *interface_server= NULL;
	struct sockaddr_in local;
    WSADATA wsaData;

	/* Initialize Server */
    /* Parse arguments */
    if (argc >1) {
        for(i=1;i <argc;i++) {
            if ( (argv[i][0] == '-') || (argv[i][0] == '/') ) {
                switch(tolower(argv[i][1])) {
                    case 'p':
                        if (!stricmp(argv[i+1], "TCP") )
                            socket_type = SOCK_STREAM;
                        else if (!stricmp(argv[i+1], "UDP") )
                            socket_type = SOCK_DGRAM;
                        else
                            Usage(argv[0]);
                        i++;
                        break;

                    case 'i':
                        interface_server = argv[++i];
                        break;
                    case 'e':
                        port = atoi(argv[++i]);
                        break;
                    default:
                        Usage(argv[0]);
                        break;
                }
            }
            else
                Usage(argv[0]);
        }
    }
    
    if ((retval = WSAStartup(0x202,&wsaData)) != 0) {
        fprintf(stderr,"WSAStartup failed with error %d\n",retval);
        WSACleanup();
        return -1;
    }
    
    if (port == 0){
        Usage(argv[0]);
    }

    local.sin_family = AF_INET;
    local.sin_addr.s_addr = (!interface_server)?INADDR_ANY:inet_addr(interface_server); 

    /* 
     * Port MUST be in Network Byte Order
     */
    local.sin_port = htons(port);

    listen_socket = socket(AF_INET, socket_type,0); // UDP socket
    
    if (listen_socket == INVALID_SOCKET){
        fprintf(stderr,"socket() failed with error %d\n",WSAGetLastError());
        WSACleanup();
        return -1;
    }
    //
    // bind() associates a local address and port combination with the
    // socket just created. This is most useful when the application is a 
    // server that has a well-known port that clients know about in advance.
    //

    if (bind(listen_socket,(struct sockaddr*)&local,sizeof(local) ) == SOCKET_ERROR) {
        fprintf(stderr,"bind() failed with error %d\n",WSAGetLastError());
        WSACleanup();
        return -1;
    }

    //
    // So far, everything we did was applicable to TCP as well as UDP.
    // However, there are certain steps that do not work when the server is
    // using UDP.
    //

    // We cannot listen() on a UDP socket.

    if (socket_type != SOCK_DGRAM) {
        if (listen(listen_socket,5) == SOCKET_ERROR) {
            fprintf(stderr,"listen() failed with error %d\n",WSAGetLastError());
            WSACleanup();
            return -1;
        }
    }
    printf("%s: 'Listening' on port %d, protocol %s\n",argv[0],port,
        (socket_type == SOCK_STREAM)?"TCP":"UDP");
}


bool info_handler(Result result, char Buffer[128]) {
	//bool shutting_down = false;
	// Set up movement parameters.
	double veloc = 30.0;
	double accel = 15.0;
	double turn_veloc = 0.0;
	double turn_accel = 0.0;
	/* Handle input from client */

	if (Buffer[0] == 'o')
	{
		wsprintf(Buffer,"PRINT IR VALUES HERE");
	}
	if (Buffer[0] == 'i')
	{
		//cout << "Forward"; 
		wsprintf(Buffer,"Forward");
		veloc = 10;
		accel = 50;
		//turn_veloc = 0;
		//turn_accel = 0;
	}
	if (Buffer[0] == 'k')
	{
		wsprintf(Buffer,"Backward");
		veloc = -10;
		accel = -50;
		//turn_veloc = 0;
		//turn_accel = 0;
		
	}
	if (Buffer[0] == 'l')
	{
		wsprintf(Buffer,"Left");
		if (turn_veloc < 0) {
			cout << "Stopping: "; 
			result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
		}
		//cout << "Left 1 second: "; 
		turn_veloc = M_PI/6.0;
		turn_accel = M_PI/2.0;
		//veloc = 0;
	}
	if (Buffer[0] == 'j')
	{
		wsprintf(Buffer,"Right");
		if (turn_veloc > 0) {
			cout << "Stopping: "; 
			result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
		}
		//std::cout << "Right 1 second: "; 
		turn_veloc = -M_PI/6.0;
		turn_accel = M_PI/2.0;
		//veloc = 0;
	}
	if (Buffer[0] == 'z')
	{
		wsprintf(Buffer,"Stopping");
		//std::cout << "Stopping "; 
		turn_veloc = 0;
		turn_accel = 0;	
		accel = -1 * veloc;
		veloc = 0;


		result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
	}
 	if (Buffer[0] == 's')
	{
		wsprintf(Buffer,"Losing control....now!");
		shutting_down = true;
	}

	result = drive->move_and_turn(0, veloc, accel, turn_veloc, turn_accel);
	CheckResult(result == Evolution::RESULT_SUCCESS);
	//result = drive->stop(0, 0, 0, Evolution::IMotorCommand::STOP_ABRUPT);
	//Evolution::Platform::millisecond_sleep (500);
	//CheckResult(result == Evolution::RESULT_SUCCESS);

	return shutting_down;
}
int main(int argc, char **argv) {

    char Buffer[128];
    int retval, fromlen, i;
	
	// server variables
	char *interface_server= NULL;
    unsigned short port=DEFAULT_PORT;
    SOCKET msgsock;

	// set up containers for motors
	Result result = Evolution::RESULT_SUCCESS;
	ResourceManager* manager = new ResourceManager (NULL, result);

	// Set up logging.
    ERSP_LOG_SET_ROOT_PRIORITY(LOG_INFO);

    signal (SIGINT, shutdown_interrupt);
#if !defined(EVOLUTION_PLATFORM_WIN32)
    signal(SIGCHLD, shutdown_interrupt);
    signal(SIGPIPE, shutdown_interrupt);
#endif

	if (evolution_init(argv, argv, result, manager) == 1)
		return 1;
	if (server_init(argc, argv, port) == -1)
		return -1;

	// main listening loop
	while(1) {
        fromlen =sizeof(from);
        //
        // accept() doesn't make sense on UDP, since we do not listen()
        //
        if (socket_type != SOCK_DGRAM) {
            msgsock = accept(listen_socket,(struct sockaddr*)&from, &fromlen);
            if (msgsock == INVALID_SOCKET) {
                fprintf(stderr,"accept() error %d\n",WSAGetLastError());
                WSACleanup();
                return -1;
            }
            printf("accepted connection from %s, port %d\n", 
                        inet_ntoa(from.sin_addr),
                        htons(from.sin_port)) ; 
        }
        else
			msgsock = listen_socket;

        // In the case of SOCK_STREAM, the server can do recv() and 
        // send() on the accepted socket and then close it.

        // However, for SOCK_DGRAM (UDP), the server will do
        // recvfrom() and sendto()  in a loop.

        if (socket_type != SOCK_DGRAM)
            retval = recv(msgsock,Buffer,sizeof (Buffer),0 );
        else {
            retval = recvfrom(msgsock,Buffer,sizeof (Buffer),0,
                (struct sockaddr *)&from,&fromlen);
            printf("Received datagram from %s\n",inet_ntoa(from.sin_addr));
        }
            
        if (retval == SOCKET_ERROR) {
            fprintf(stderr,"recv() failed: error %d\n",WSAGetLastError());
            closesocket(msgsock);
            continue;
        }
        if (retval == 0) {
            printf("Client closed connection\n");
            closesocket(msgsock);
            continue;
        }
        printf("Received %d bytes, data [%s] from client\n",retval,Buffer);

		// send buffer to info_handler to do stuff with it
		// if info_handler returns true, time to shut down system
		if (info_handler(result, Buffer) == true)
		{
			std::cout << "Deactivating resources\n";

			// Release allocated interfaces
			if (resource_container && drive) {
				resource_container->release_interface (0, drive);
				drive = NULL;
			}
	  	    if (resource_container && sensors) {
				for (unsigned int i = 0; i < sensors->size(); i++)
					resource_container->release_interface (0, (*sensors)[i]);
				sensors = NULL;
		    // Delete string arrays
			if (names) {
				for (unsigned int i = 0; i < names->size(); i++)
					delete [] (*names)[i];
    }
    }
		}
		/* server continues */
		double v, w;
		printf("Echoing same data back to client\n");
        if (socket_type != SOCK_DGRAM)
            retval = send(msgsock,Buffer,sizeof(Buffer),0);
        else
            retval = sendto(msgsock,Buffer,sizeof (Buffer),0,
                (struct sockaddr *)&from,fromlen);
        if (retval == SOCKET_ERROR) {
            fprintf(stderr,"send() failed: error %d\n",WSAGetLastError());
        }
        if (socket_type != SOCK_DGRAM){
            printf("Terminating connection\n");
            closesocket(msgsock);
        }
        else 
            printf("UDP server looping back for more requests\n");
        continue;

		// Deactivate drive system.
		
		}
	return (0);
}
